"""The object returned by the `--settings` option."""
from typing import Dict

import attrs

from civ4save.enums import vanilla as e


@attrs.define(slots=True)
class Settings:
    """Class representing the game's settings (sp only)."""
    # CvInitCore
    game_type: e.GameType
    game_name: str
    map_script: str
    """Name of the map script"""
    world_size: e.WorldType
    climate: e.ClimateType
    sea_level: e.SeaLevelType
    start_era: e.EraType
    game_speed: e.GameSpeedType
    game_options: Dict[str, bool]
    """Each GameOption and whether it's enabled"""
    max_turns: int
    advanced_start_points: int
    victories: Dict[str, bool]
    """Each VictoryType and whether it's enabled"""
    num_civs: int
    """Number of civs (not counting barbs)"""

    # CvGame
    start_turn: int
    start_year: int
    handicap: e.HandicapType
    map_random_seed: int
    soren_random_seed: int
    culture_victory_cities: int
    culture_victory_level: e.CultureLevelType

    # CvMap
    grid_width: int
    grid_height: int
    wrap_x: bool
    wrap_y: bool

    @classmethod
    def from_struct(cls, cv_init, cv_game, cv_map):
        """Return `Settings` from the cv_init, cv_game, and cv_map parsed structs."""
        game_type = e.GameType[cv_init.game_type]
        game_speed = e.GameSpeedType[cv_init.game_speed]
        world_size = e.WorldType[cv_init.world_size]
        climate = e.ClimateType[cv_init.climate]
        sea_level = e.SeaLevelType[cv_init.sea_level]
        start_era = e.EraType[cv_init.start_era]

        handicap = e.HandicapType[cv_game.handicap]
        culture_victory_level = e.CultureLevelType[cv_game.culture_victory_level]

        game_options = {
            e.GameOptionType(n).name: v for n, v in enumerate(cv_init.game_options)
        }

        advanced_start_points = 0
        if game_options["GAMEOPTION_ADVANCED_START"]:
            advanced_start_points = cv_init.advanced_start_points

        victories = {e.VictoryType(n).name: v for n, v in enumerate(cv_init.victories)}

        num_civs = len([c for c in cv_init.civs[:-1] if c != "NO_CIVILIZATION"])

        return cls(
            game_type=game_type,
            game_name=cv_init.game_name,
            map_script=cv_init.map_script_name,
            world_size=world_size,
            climate=climate,
            sea_level=sea_level,
            start_era=start_era,
            game_speed=game_speed,
            game_options=game_options,
            max_turns=cv_init.max_turns,
            advanced_start_points=advanced_start_points,
            victories=victories,
            num_civs=num_civs,
            start_turn=cv_game.start_turn,
            start_year=cv_game.start_year,
            handicap=handicap,
            map_random_seed=cv_game.map_random_seed,
            soren_random_seed=cv_game.soren_random_seed,
            culture_victory_cities=cv_game.num_culture_victory_cities,
            culture_victory_level=culture_victory_level,
            grid_width=cv_map.grid_width,
            grid_height=cv_map.grid_height,
            wrap_x=cv_map.wrap_x,
            wrap_y=cv_map.wrap_y,
        )



"""This module is responsible for parsing the XML files."""
import re
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Generator, Optional

import xmltodict
from jinja2 import Template

from civ4save.utils import get_game_dir


def get_n_parents(path: Path, num: int) -> Path:
    """Return a shortened path with only `num` parents."""
    path = path.resolve()
    name = path.name
    parents = []
    for n in range(num):
        path = path.parent
        if not path.name:
            break
        parents.append(path.name)
    return Path(parents[-1]).joinpath(*reversed(parents[:-1])) / name


@dataclass
class Civ4XmlFile:
    """Helper class for transforming XML file to Python Enum."""
    path: Path
    _parsed: dict = field(default_factory=dict)

    def __post_init__(self):
        """Validate actually xml file."""
        if not self.path.suffix == ".xml":
            raise ValueError(f"{get_n_parents(self.path, 4)} not a xml file")

    def __str__(self) -> str:
        """Return shorted string representation of XML file path."""
        return str(get_n_parents(self.path, 4))

    @property
    def name(self) -> str:
        """The file name."""
        return self.path.name

    @property
    def enum_name(self) -> str:
        """Name of the Enum corresponding to the XML file."""
        name = self.path.stem.replace("Infos", "")
        name = name.replace("CIV4", "")
        name = name.replace("Info", "")
        name = name.replace("_", "")
        return name + "Type"

    @property
    def negative_one(self) -> str:
        """Return the default NO_<SOMETHING> = -1 name for the Enum."""
        name = re.sub(r'(?<!^)(?=[A-Z])', '_', self.enum_name)
        return "NO_" + name.replace("_Type", "").upper()

    @property
    def parent_category(self) -> Optional[str]:
        """Example GameInfo/CIV4EraInfos.xml the category is GameInfo."""
        parent = self.path.parent.name
        if self.path.parent.name == "XML":
            return None
        return parent

    @property
    def root_key(self) -> str:
        """Top level key."""
        if not self._parsed:
            self._read()
        return [key for key in self._parsed].pop()

    def types(self):
        """These become the Enum's members."""
        root = self._parsed[self.root_key]
        key = list(root.keys())[-1]
        names = [(self.negative_one, -1)]
        for n, entry in enumerate(root[key][key[:-1]]):
            names.append((entry["Type"], n))
        return names

    def to_enum(self):
        """Return and Enum derived from the XML file."""
        return Enum(self.enum_name, self.types())

    def _read(self) -> dict:
        if not self.path.exists():
            raise FileNotFoundError(f"{self.path} does not exist!")
        with open(self.path, "r") as xml_file:
            self._parsed = xmltodict.parse(xml_file.read())
        return self._parsed


def xml_files_iter() -> Generator[Civ4XmlFile, None, None]:
    """Generator yielding every XML file that isn't a Schema or Text type.

    BTS takes precendence over Warlords and Warlords over vanilla.

    Example:
        if both <BTS path>/CIV4BuildingInfos.xml, <Vanilla path>/CIV4BuildingInfos.xml
            only yield the BTS one
    """
    # TODO replace the ugliness with a glob** of some kind
    game_dir = get_game_dir()
    vanilla_xml = game_dir / "Assets" / "XML"
    warlords_xml = game_dir / "Warlords" / "Assets" / "XML"
    bts_xml = game_dir / "Beyond the Sword" / "Assets" / "XML"

    seen = set()
    for xml_dir in [bts_xml, warlords_xml, vanilla_xml]:
        if not xml_dir.exists():
            continue
        for file_or_dir in xml_dir.iterdir():
            if file_or_dir.is_dir():
                for file in file_or_dir.iterdir():
                    if (
                        "Schema" not in file.name
                        and "Text" not in file.name
                        and file.name not in seen
                    ):
                        seen.add(file.name)
                        yield Civ4XmlFile(file)
            else:
                if (
                    "Schema" not in file_or_dir.name
                    and "Text" not in file_or_dir.name
                    and file_or_dir.name not in seen
                ):
                    seen.add(file_or_dir.name)
                    yield Civ4XmlFile(file_or_dir)


def write_out_enum(e):
    """Write Enum to stdout using a jinja template."""
    data = {
        "enum_name": e.__name__,
        "members": [{"name": m, "value": e[m].value} for m in e.__members__],
    }
    template = """
class {{ enum_name }}(Enum):
    {%- for m in members %}
    {{ m.name }} = {{ m.value }}
    {%- endfor -%}
{{ '\n' }}
"""

    j2_template = Template(template)
    print(j2_template.render(data))


def make_enums() -> None:
    """Write all enums to stdout."""
    print("from enum import Enum")
    print()
    for xml_file in xml_files_iter():
        try:
            xml_file._read()
        except UnicodeDecodeError:
            # FIXME some files are not unicode encoded
            continue
        try:
            write_out_enum(xml_file.to_enum())
        except Exception:
            continue



"""Matches the byte structure of a saved game file but in managable chunks.

Everything is little endian bc x86.

Todo:
    Could define string types as?
        W_STRING = partial(C.PaddedString, encoding="utf_16_le")
        STRING = partial(C.PaddedString, encoding="utf_8")
"""
import construct as C

from .objects import Context
from .utils import get_enum_length

NUM_YIELD_TYPES = 3

# Type Aliases
INT = C.Int32sl
UINT = C.Int32ul
SHORT = C.Int16sl
USHORT = C.Int16ul
CHAR = C.Int8sl
UCHAR = C.Int8ul


def metadata():
    """First few bytes containing version and zlib info."""
    return C.Struct(
        "version" / INT,
        "save_bits" / C.Array(8, INT),
        "bytes_to_zlib_magic_number" / INT,
        "_idx" / C.Tell,
    )


def cv_init_core(ctx: Context):
    """As defined in CvInitCore."""
    return C.Struct(
        "save_flag" / UINT,
        "game_type" / C.Enum(INT, ctx.enums.GameType),
        "sz_game_name" / INT,
        "game_name"
        / C.PaddedString(C.this.sz_game_name * 2, "utf_16_le"),  # w_char hack
        "sz_game_password" / UINT,
        "game_password" / C.PaddedString(C.this.sz_game_password * 2, "utf_16_le"),
        "sz_admin_password" / UINT,
        "admin_password" / C.PaddedString(C.this.sz_admin_password * 2, "utf_16_le"),
        "sz_map_script_name" / UINT,
        "map_script_name" / C.PaddedString(C.this.sz_map_script_name * 2, "utf_16_le"),
        "wb_map_no_players" / C.Flag,
        "world_size" / C.Enum(INT, ctx.enums.WorldType),
        "climate" / C.Enum(INT, ctx.enums.ClimateType),
        "sea_level" / C.Enum(INT, ctx.enums.SeaLevelType),
        "start_era" / C.Enum(INT, ctx.enums.EraType),
        "game_speed" / C.Enum(INT, ctx.enums.GameSpeedType),
        "turn_timer" / INT,
        "calendar" / INT,
        "num_custom_map_options" / INT,
        "num_hidden_custom_map_options" / INT,
        "custom_map_options" / INT[C.this.num_custom_map_options],
        "num_victories" / INT,
        "victories" / C.Flag[C.this.num_victories],
        "game_options" / C.Flag[get_enum_length(ctx.enums.GameOptionType)],
        "mp_game_options" / C.Flag[get_enum_length(ctx.enums.MultiplayerOptionType)],
        "stat_reporting" / C.Flag,
        "game_turn" / INT,
        "max_turns" / INT,
        "pitboss_turn_time" / INT,
        "target_score" / INT,
        "max_city_eliminations" / INT,
        "advanced_start_points" / INT,
        "leader_names"
        / C.Array(
            ctx.max_players,
            C.Struct(
                "sz" / UINT,
                "name" / C.PaddedString(C.this.sz * 2, "utf_16_le"),
            ),
        ),
        "civ_descriptions"
        / C.Array(
            ctx.max_players,
            C.Struct(
                "sz" / UINT,
                "description" / C.PaddedString(C.this.sz * 2, "utf_16_le"),
            ),
        ),
        "civ_short_descriptions"
        / C.Array(
            ctx.max_players,
            C.Struct(
                "sz" / UINT,
                "short_description" / C.PaddedString(C.this.sz * 2, "utf_16_le"),
            ),
        ),
        "civ_adjectives"
        / C.Array(
            ctx.max_players,
            C.Struct(
                "sz" / UINT,
                "adjective" / C.PaddedString(C.this.sz * 2, "utf_16_le"),
            ),
        ),
        "emails"
        / C.Array(
            ctx.max_players,
            C.Struct(
                "sz" / UINT,
                "email" / C.PaddedString(C.this.sz, "utf_8"),
            ),
        ),
        "smtp_hosts"
        / C.Array(
            ctx.max_players,
            C.Struct(
                "sz" / UINT,
                "smtp_host" / C.PaddedString(C.this.sz, "utf_8"),
            ),
        ),
        "white_flags" / C.Flag[ctx.max_players],
        "mystery?" / INT[ctx.max_players],  # TODO: What is this? Padding?
        "flag_decals"
        / C.Array(
            ctx.max_players,
            C.Struct(
                "sz" / UINT,
                "flag_decal" / C.PaddedString(C.this.sz * 2, "utf_16_le"),
            ),
        ),
        "civs"
        / C.Array(
            ctx.max_players,
            C.Enum(INT, ctx.enums.CivilizationType),
        ),
        "leaders"
        / C.Array(
            ctx.max_players,
            C.Enum(INT, ctx.enums.LeaderHeadType),
        ),
        "teams" / INT[ctx.max_players],
        "handicaps"
        / C.Array(
            ctx.max_players,
            C.Enum(INT, ctx.enums.HandicapType),
        ),
        "colors" / INT[ctx.max_players],
        "art_style" / INT[ctx.max_players],
        "slot_statuses" / INT[ctx.max_players],
        "slot_claims" / INT[ctx.max_players],
        "playable_civs" / C.Flag[ctx.max_players],
        "minor_nation_civs" / C.Flag[ctx.max_players],
        "_idx" / C.Tell,
    )


def trade_data(ctx: Context):
    """Used in cv_game."""
    return C.Struct(
        "item" / C.Enum(INT, ctx.enums.TradeableItem),
        "extra_data" / INT,  # BonusType sometimes applicable
        "offering" / C.Flag,
        C.Padding(1),
        "hidden" / C.Flag,
        C.Padding(1),
    )


def vote_option(ctx: Context):
    """Used in cv_game."""
    return C.Struct(
        "type" / C.Enum(INT, ctx.enums.VoteType),
        "player" / INT,
        "city_id" / INT,
        "other_player" / INT,
        "sz_text" / INT,
        "text" / C.PaddedString(C.this.sz_text * 2, "utf_16_le"),
    )


def cv_game(ctx: Context):
    """As defined in CvGameAi + CvGame."""
    TradeData = trade_data(ctx)
    VoteOption = vote_option(ctx)
    return C.Struct(
        "game_ai_ui_flag" / UINT,
        "game_ai_pad" / INT,
        "game_ui_flag" / UINT,
        "elapsed_game_turns" / INT,
        "start_turn" / INT,
        "start_year" / INT,
        "estimated_end_turn" / INT,
        "turn_slice" / INT,
        "cutoff_slice" / INT,
        "num_game_turn_active" / INT,
        "total_cities" / INT,
        "total_population" / INT,
        "trade_routes" / INT,
        "free_trade_count" / INT,
        "no_nukes_count" / INT,
        "nukes_exploded" / INT,
        "max_population" / INT,
        "max_land" / INT,
        "max_tech" / INT,
        "max_wonders" / INT,
        "init_population" / INT,
        "init_land" / INT,
        "init_tech" / INT,
        "init_wonders" / INT,
        "ai_autoplay" / INT,
        "score_dirty" / C.Flag,
        "circumnavigated" / C.Flag,
        "final_initialized" / C.Flag,
        "hot_pbem_between_turns" / C.Flag,
        "nukes_valid" / C.Flag,
        "handicap" / C.Enum(INT, ctx.enums.HandicapType),
        "pause_player" / INT,  # -1 = NO_PLAYER, otherwise is index of players array
        "best_land_unit" / C.Enum(INT, ctx.enums.UnitType),
        "winner" / INT,  # -1 = NO_TEAM, otherwise is index of teams array
        "victory" / C.Enum(INT, ctx.enums.VictoryType),
        "game_state" / C.Enum(INT, ctx.enums.GameStateType),
        "sz_script_data" / UINT,
        "script_data" / C.PaddedString(C.this.sz_script_data, "utf_8"),
        "ai_rank_player" / INT[ctx.max_players],
        "ai_player_rank" / INT[ctx.max_players],
        "ai_player_score" / INT[ctx.max_players],
        "ai_rank_team" / INT[ctx.max_players],
        "ai_team_rank" / INT[ctx.max_players],
        "ai_team_score" / INT[ctx.max_players],
        "unit_created_counts" / INT[get_enum_length(ctx.enums.UnitType)],
        "unit_class_created_counts" / INT[get_enum_length(ctx.enums.UnitClassType)],
        "building_class_created_counts"
        / INT[get_enum_length(ctx.enums.BuildingClassType)],
        "project_created_counts" / INT[get_enum_length(ctx.enums.ProjectType)],
        "force_civic_counts" / INT[get_enum_length(ctx.enums.CivicType)],
        "vote_outcomes" / INT[get_enum_length(ctx.enums.VoteType)],
        "religion_game_turn_founded" / INT[get_enum_length(ctx.enums.ReligionType)],
        "corporation_game_turn_founded"
        / INT[get_enum_length(ctx.enums.CorporationType)],
        "secretary_general_timer" / INT[get_enum_length(ctx.enums.VoteSourceType)],
        "vote_timer" / INT[get_enum_length(ctx.enums.VoteSourceType)],
        "diplo_vote" / INT[get_enum_length(ctx.enums.VoteSourceType)],
        "special_unit_valid" / C.Flag[get_enum_length(ctx.enums.SpecialUnitType)],
        "special_building_valid"
        / C.Flag[get_enum_length(ctx.enums.SpecialBuildingType)],
        "religion_slot_taken" / C.Flag[get_enum_length(ctx.enums.ReligionType)],
        "holy_cities"
        / C.Array(
            get_enum_length(ctx.enums.ReligionType),
            C.Struct("owner" / INT, "city_id?" / INT),
        ),
        "corporation_headquarters"
        / C.Array(
            get_enum_length(ctx.enums.CorporationType),
            C.Struct("owner" / INT, "city_id?" / INT),
        ),
        "num_cities_destroyed" / INT,
        "cities_destroyed"
        / C.Array(
            C.this.num_cities_destroyed,
            C.Struct(
                "sz" / UINT,
                "name" / C.PaddedString(C.this.sz * 2, "utf_16_le"),
            ),
        ),
        "num_great_people_born" / INT,
        "great_people_born"
        / C.Array(
            C.this.num_great_people_born,
            C.Struct(
                "sz" / UINT,
                "name" / C.PaddedString(C.this.sz * 2, "utf_16_le"),
            ),
        ),
        "deals_num_slots" / INT,
        "deals_last_index" / INT,
        "deals_free_list_head" / INT,
        "deals_free_list_count" / INT,
        "deals_current_id" / INT,
        "deals_next_free_index_array" / INT[C.this.deals_num_slots],
        "sz_deals" / INT,
        "deals"
        / C.Array(
            C.this.sz_deals,
            C.Struct(
                "ui_flag" / UINT,
                "id" / INT,
                "initial_game_turn" / INT,
                "first_player" / INT,
                "second_player" / INT,
                "sz_first_trades" / INT,
                "first_trades" / C.Array(C.this.sz_first_trades, TradeData),
                "sz_second_trades" / INT,
                "second_trades" / C.Array(C.this.sz_second_trades, TradeData),
            ),
        ),
        "vote_selections_num_slots" / INT,
        "vote_selections_last_index" / INT,
        "vote_selections_free_list_head" / INT,
        "vote_selections_free_list_count" / INT,
        "vote_selections_current_id" / INT,
        "vote_selections_next_free_index_array" / INT[C.this.vote_selections_num_slots],
        "sz_vote_selections" / INT,
        "vote_selections"
        / C.Array(
            C.this.sz_vote_selections,
            C.Struct(
                "vote_id" / INT,
                "vote_source" / C.Enum(INT, ctx.enums.VoteSourceType),
                "sz_vote_options" / INT,
                "vote_options" / C.Array(C.this.sz_vote_options, VoteOption),
            ),
        ),
        "votes_triggered_num_slots" / INT,
        "votes_triggered_last_index" / INT,
        "votes_triggered_free_list_head" / INT,
        "votes_triggered_free_list_count" / INT,
        "votes_triggered_current_id" / INT,
        "votes_triggered_next_free_index_array" / INT[C.this.vote_selections_num_slots],
        "sz_votes_triggered" / INT,
        "votes_triggered"
        / C.Array(
            C.this.sz_votes_triggered,
            C.Struct(
                "vote_id" / INT,
                "vote_source" / C.Enum(INT, ctx.enums.VoteSourceType),
                "vote_option" / VoteOption,
            ),
        ),
        "map_random_seed" / UINT,
        "soren_random_seed" / UINT,
        "sz_replay_messages" / INT,
        "replay_messages"
        / C.Array(
            C.this.sz_replay_messages,
            C.Struct(
                "turn" / INT,
                "type" / C.Enum(INT, ctx.enums.ReplayMessageType),
                "plot_x" / INT,
                "plot_y" / INT,
                "player" / INT,
                "sz_text" / INT,
                "text" / C.PaddedString(C.this.sz_text * 2, "utf_16_le"),
                "e_color" / INT,
            ),
        ),
        "num_sessions" / INT,
        "sz_plot_extra_yields" / INT,
        "plot_extra_yields"
        / C.Array(
            C.this.sz_plot_extra_yields,
            C.Struct(
                "plot_x" / INT,
                "plot_y" / INT,
                "extra_yields" / INT[NUM_YIELD_TYPES],
            ),
        ),
        "sz_plot_extra_costs" / INT,
        "plot_extra_costs"
        / C.Array(
            C.this.sz_plot_extra_costs,
            C.Struct(
                "plot_x" / INT,
                "plot_y" / INT,
                "extra_costs" / INT[NUM_YIELD_TYPES],
            ),
        ),
        "sz_vote_source_religions" / INT,
        "vote_source_religions"
        / C.Array(
            C.this.sz_vote_source_religions,
            C.Struct(
                "vote_source" / C.Enum(INT, ctx.enums.VoteSourceType),
                "religion" / C.Enum(INT, ctx.enums.ReligionType),
            ),
        ),
        "sz_inactive_triggers" / INT,
        "inactive_triggers"
        / C.Array(
            C.this.sz_inactive_triggers,
            C.Enum(INT, ctx.enums.EventTriggerType),
        ),
        "shrine_building_count" / INT,
        "shrine_buildings"
        / C.Array(
            get_enum_length(ctx.enums.BuildingType),
            C.Enum(INT, ctx.enums.BuildingType),
        ),
        "shrine_religion"
        / C.Array(
            get_enum_length(ctx.enums.BuildingType),
            C.Enum(INT, ctx.enums.BuildingType),
        ),
        "num_culture_victory_cities" / INT,
        "culture_victory_level" / C.Enum(INT, ctx.enums.CultureLevelType),
        "_idx" / C.Tell,
    )


def cv_map_base(ctx: Context):
    """As defined in CvMap. Plots are treated separately."""
    return C.Struct(
        "map_ui_flag" / UINT,
        "map_unknown_ints" / CHAR[8],
        "grid_width" / INT,
        "grid_height" / INT,
        "land_plots" / INT,
        "owned_plots" / INT,
        "top_latitude" / INT,
        "bottom_latitude" / INT,
        "next_river_id" / INT,
        "wrap_x" / C.Flag,
        "wrap_y" / C.Flag,
        "bonus_counts" / INT[get_enum_length(ctx.enums.BonusType)],
        "bonus_counts_on_land" / INT[get_enum_length(ctx.enums.BonusType)],
        "_idx" / C.Tell,
    )


def cv_plot(ctx: Context):
    """As defined in CvPlot (called in CvMap)."""
    return C.Struct(
        "ui_flag" / UINT,
        "x" / SHORT,
        "y" / SHORT,
        "area_id" / INT,
        "feature_variety" / SHORT,
        "ownership_duration" / SHORT,
        "improvement_duration" / SHORT,
        "upgrade_progress" / SHORT,
        "force_unowned_timer" / SHORT,
        "city_radius_count" / SHORT,
        "river_id" / INT,
        "min_original_start_distance" / SHORT,
        "recon_count" / SHORT,
        "river_crossing_count" / SHORT,
        "starting_plot" / C.Flag,
        "hills" / C.Flag,
        "north_of_river" / C.Flag,
        "west_of_river" / C.Flag,
        "irrigated" / C.Flag,
        "potential_city_work" / C.Flag,
        "owner" / CHAR,
        "plot_type" / C.Enum(SHORT, ctx.enums.PlotType),
        "terrain_type" / C.Enum(SHORT, ctx.enums.TerrainType),
        "feature_type" / C.Enum(SHORT, ctx.enums.FeatureType),
        "bonus_type" / C.Enum(SHORT, ctx.enums.BonusType),
        "improvement_type" / C.Enum(SHORT, ctx.enums.ImprovementType),
        "route_type" / SHORT,
        "river_north_south" / CHAR,
        "river_east_west" / CHAR,
        "plot_city_owner" / INT,
        "plot_city_id" / INT,
        "working_city_owner" / INT,
        "working_city_id" / INT,
        "working_city_override_owner" / INT,
        "working_city_override_id" / INT,
        "yields" / SHORT[NUM_YIELD_TYPES],
        "sz_culture" / CHAR,
        "culture" / INT[C.this.sz_culture],
        "sz_found_value" / CHAR,
        "found_value" / SHORT[C.this.sz_found_value],
        "sz_player_city_radius" / CHAR,
        "player_city_radius" / CHAR[C.this.sz_player_city_radius],
        "sz_plot_group" / CHAR,
        "plot_group" / INT[C.this.sz_plot_group],
        "sz_visibility" / CHAR,
        "visibility" / SHORT[C.this.sz_visibility],
        "sz_stolen_visibility" / CHAR,
        "stolen_visibility" / SHORT[C.this.sz_stolen_visibility],
        "sz_blockaded" / CHAR,
        "blockaded" / SHORT[C.this.sz_blockaded],
        "sz_revealed_owner" / CHAR,
        "revealed_owner" / CHAR[C.this.sz_revealed_owner],
        "sz_direction_types" / CHAR,
        "river_crossings" / C.Flag[C.this.sz_direction_types],
        "sz_revealed" / CHAR,
        "revealed" / C.Flag[C.this.sz_revealed],
        "sz_revealed_improvement_type" / CHAR,
        "revealed_improvement_type" / SHORT[C.this.sz_revealed_improvement_type],
        "sz_revealed_route_type" / CHAR,
        "revealed_route_type" / SHORT[C.this.sz_revealed_route_type],
        "sz_plot_script_data" / INT,
        "plot_script_data" / C.PaddedString(C.this.sz_plot_script_data, "utf_8"),
        "sz_build_progress" / INT,
        "build_progress" / SHORT[C.this.sz_build_progress],
        "sz_culture_range_cities" / CHAR,
        "culture_range_cities"
        / C.Array(
            C.this.sz_culture_range_cities,
            C.Struct(
                "crc_sz" / INT,
                "crc" / CHAR[C.this.crc_sz],
            ),
        ),
        "sz_invisible_visibility" / CHAR,
        "invisible_visibles"
        / C.Array(
            C.this.sz_invisible_visibility,
            C.Struct(
                "inv_sz" / INT,
                "inv_vis" / SHORT[C.this.inv_sz],
            ),
        ),
        "sz_units" / INT,
        "units"
        / C.Array(
            C.this.sz_units,
            C.Struct("owner" / INT, "id" / INT),
        ),
        "end_plot_index" / C.Tell,
    )


def cv_plot_debug(ctx: Context):
    """Can play with this for debugging without messing up actual cv_plot struct."""
    return C.Struct(
        "ui_flag" / UINT,
        "x" / SHORT,
        "y" / SHORT,
        "area_id" / INT,
        "feature_variety" / SHORT,
        "ownership_duration" / SHORT,
        "improvement_duration" / SHORT,
        "upgrade_progress" / SHORT,
        "force_unowned_timer" / SHORT,
        "city_radius_count" / SHORT,
        "river_id" / INT,
        "min_original_start_distance" / SHORT,
        "recon_count" / SHORT,
        "river_crossing_count" / SHORT,
        "starting_plot" / C.Flag,
        "hills" / C.Flag,
        "north_of_river" / C.Flag,
        "west_of_river" / C.Flag,
        "irrigated" / C.Flag,
        "potential_city_work" / C.Flag,
        "owner" / CHAR,
        "plot_type" / C.Enum(SHORT, ctx.enums.PlotType),
        "terrain_type" / C.Enum(SHORT, ctx.enums.TerrainType),
        "feature_type" / C.Enum(SHORT, ctx.enums.FeatureType),
        "bonus_type" / C.Enum(SHORT, ctx.enums.BonusType),
        "improvement_type" / C.Enum(SHORT, ctx.enums.ImprovementType),
        "route_type" / SHORT,
        "river_north_south" / CHAR,
        "river_east_west" / CHAR,
        "plot_city_owner" / INT,
        "plot_city_id" / INT,
        "working_city_owner" / INT,
        "working_city_id" / INT,
        "working_city_override_owner" / INT,
        "working_city_override_id" / INT,
        "yields" / SHORT[NUM_YIELD_TYPES],  # dutch
        "sz_culture" / CHAR,
        "culture" / INT[C.this.sz_culture],
        "sz_found_value" / CHAR,
        "_idx" / C.Tell,
        # "found_value" / SHORT[C.this.sz_found_value],
        # "sz_player_city_radius" / CHAR,
        # "player_city_radius" / CHAR[C.this.sz_player_city_radius],
        # "sz_plot_group" / CHAR,
        # "plot_group" / INT[C.this.sz_plot_group],
        # "sz_visibility" / CHAR,
        # "visibility" / SHORT[C.this.sz_visibility],
        # "sz_stolen_visibility" / CHAR,
        # "stolen_visibility" / SHORT[C.this.sz_stolen_visibility],
        # "sz_blockaded" / CHAR,
        # "blockaded" / SHORT[C.this.sz_blockaded],
        # "sz_revealed_owner" / CHAR,
        # "revealed_owner" / CHAR[C.this.sz_revealed_owner],
        # "sz_direction_types" / CHAR,
        # "river_crossings" / C.Flag[C.this.sz_direction_types],
        # "sz_revealed" / CHAR,
        # "revealed" / C.Flag[C.this.sz_revealed],
        # "sz_revealed_improvement_type" / CHAR,
        # "revealed_improvement_type" / SHORT[C.this.sz_revealed_improvement_type],
        # "sz_revealed_route_type" / CHAR,
        # "revealed_route_type" / SHORT[C.this.sz_revealed_route_type],
        # "sz_plot_script_data" / INT,
        # "plot_script_data" / C.PaddedString(C.this.sz_plot_script_data, "utf_8"),
        # "sz_build_progress" / INT,
        # "build_progress" / SHORT[C.this.sz_build_progress],
    )


def cv_map_area(ctx: Context):
    """As defined in CvArea (called in CvMap).

    UNDER CONSTRUCTION
    """
    return C.Struct(
        #   pStream->Write(uiFlag);		// flag for expansion
        "area_ui_flag" / UINT,
        # 	pStream->Write(m_iID);
        "area_id" / INT,
        # 	pStream->Write(m_iNumTiles);
        "num_tiles" / INT,
        # 	pStream->Write(m_iNumOwnedTiles);
        "num_owned_tiles" / INT,
        # 	pStream->Write(m_iNumRiverEdges);
        "num_river_edges" / INT,
        # 	pStream->Write(m_iNumUnits);
        "num_units" / INT,
        # 	pStream->Write(m_iNumCities);
        "num_cities" / INT,
        # 	pStream->Write(m_iTotalPopulation);
        "total_population" / INT,
        # 	pStream->Write(m_iNumStartingPlots);
        "num_starting_plots" / INT,
        # 	pStream->Write(m_bWater);
        "water" / C.Flag,
        # 	pStream->Write(MAX_PLAYERS, m_aiUnitsPerPlayer);
        "units_per_player" / C.Array(ctx.max_players, INT),
        # 	pStream->Write(MAX_PLAYERS, m_aiAnimalsPerPlayer);
        "animals_per_player" / C.Array(ctx.max_players, INT),
        # 	pStream->Write(MAX_PLAYERS, m_aiCitiesPerPlayer);
        "cities_per_player" / C.Array(ctx.max_players, INT),
        # 	pStream->Write(MAX_PLAYERS, m_aiPopulationPerPlayer);
        "population_per_player" / C.Array(ctx.max_players, INT),
        # 	pStream->Write(MAX_PLAYERS, m_aiBuildingGoodHealth);
        "building_good_health" / C.Array(ctx.max_players, INT),
        # 	pStream->Write(MAX_PLAYERS, m_aiBuildingBadHealth);
        "building_bad_health" / C.Array(ctx.max_players, INT),
        # 	pStream->Write(MAX_PLAYERS, m_aiBuildingHappiness);
        "building_happiness" / C.Array(ctx.max_players, INT),
        # 	pStream->Write(MAX_PLAYERS, m_aiFreeSpecialist);
        "free_specialist" / C.Array(ctx.max_players, INT),
        # 	pStream->Write(MAX_PLAYERS, m_aiPower);
        "power" / C.Array(ctx.max_players, INT),
        # 	pStream->Write(MAX_PLAYERS, m_aiBestFoundValue);
        "best_found_value" / C.Array(ctx.max_players, INT),
        # 	pStream->Write(MAX_TEAMS, m_aiNumRevealedTiles);
        "num_revealed_tiles" / C.Array(ctx.max_teams, INT),
        # 	pStream->Write(MAX_TEAMS, m_aiCleanPowerCount);
        "clean_power_count" / C.Array(ctx.max_teams, INT),
        # 	pStream->Write(MAX_TEAMS, m_aiBorderObstacleCount);
        "border_obstacle_count" / C.Array(ctx.max_teams, INT),
        # pStream->Read(MAX_TEAMS, (int*)m_aeAreaAIType);
        # "area_ai_type" / C.Array(ctx.max_teams, C.Enum(INT, ctx.enums.AreaAIType)),
        # for (iI=0;iI<MAX_PLAYERS;iI++)
        # {
        # 	pStream->Read((int*)&m_aTargetCities[iI].eOwner);
        # 	pStream->Read(&m_aTargetCities[iI].iID);
        # }
        "target_cities"
        / C.Array(
            ctx.max_players,
            C.Struct(
                "player_id" / INT,
                "city_id" / INT,
            ),
        ),
        # for (iI = 0; iI < MAX_PLAYERS; iI++)
        # {
        # 	pStream->Read(NUM_YIELD_TYPES, m_aaiYieldRateModifier[iI]);
        # }
        "yield_rate_modifier" / C.Array(ctx.max_players, C.Array(NUM_YIELD_TYPES, INT)),
        # for (iI = 0; iI < MAX_PLAYERS; iI++)
        # {
        # 	pStream->Read(NUM_UNITAI_TYPES, m_aaiNumTrainAIUnits[iI]);
        # }
        "num_train_ai_units" / C.Array(ctx.max_players, C.Array(41, INT)),
        # for (iI = 0; iI < MAX_PLAYERS; iI++)
        # {
        # 	pStream->Read(NUM_UNITAI_TYPES, m_aaiNumAIUnits[iI]);
        # }
        "num_ai_units" / C.Array(ctx.max_players, C.Array(41, INT)),
        # pStream->Read(GC.getNumBonusInfos(), m_paiNumBonuses);
        "num_bonuses" / INT[get_enum_length(ctx.enums.BonusType)],
        # pStream->Read(GC.getNumImprovementInfos(), m_paiNumImprovements);
        "num_improvements" / INT[get_enum_length(ctx.enums.ImprovementType)],
        "_idx" / C.Tell,
    )


def cv_map_areas(ctx: Context):
    """As defined in CvArea (CvMap).

    CvMapArea FFlist metadata plus CvArea array
    UNDER CONSTRUCTION
    """
    CvMapArea = cv_map_area(ctx)
    return C.Struct(
        "areas_num_slots" / INT,
        "areas_last_index" / INT,
        "areas_free_list_head" / INT,
        "areas_free_list_count" / INT,
        "areas_current_id" / INT,
        "areas_next_free_index_array" / INT[C.this.areas_num_slots],
        "sz_areas" / INT,
        "areas" / C.Array(C.this.sz_areas, CvMapArea),
        # "areas" / C.Array(1, CvMapArea),
        "peek" / INT[32],
        "_idx" / C.Tell,
    )



"""Some potentially interesting stuff with the Leaders."""
import json
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import importlib_resources
import xmltodict

from ..enums.vanilla import (
    CivicType,
    ImprovementType,
    LeaderHeadType,
    ReligionType,
    TraitType,
    UnitAIType,
)


@dataclass
class Leader:
    """Object holding Leader attributes."""

    type: LeaderHeadType
    description: str
    favorite_civic: CivicType
    favorite_religion: ReligionType
    traits: Tuple[TraitType, TraitType]
    flavors: Dict[str, int]
    unit_ai_weight_modifiers: Dict[str, int]
    improvement_weight_modifiers: Dict[str, int]
    wonder_construct_rand: int
    base_attitude: int
    base_peace_weight: int
    peace_weight_rand: int
    warmonger_respect: int
    espionage_weight: int
    refuse_to_talk_war_threshold: int
    no_tech_trade_threshold: int
    tech_trade_known_percent: int
    max_gold_trade_percent: int
    max_gold_trade_per_turn_percent: int
    max_war_rand: int
    max_war_nearby_power_ratio: int
    max_war_distant_power_ratio: int
    max_war_min_adjacent_land_percent: int
    limited_war_rand: int
    limited_war_power_ratio: int
    dogpile_war_rand: int
    make_peace_rand: int
    declare_war_trade_rand: int
    demand_rebuked_sneak_prob: int
    demand_rebuked_war_prob: int
    raze_city_prob: int
    build_unit_prob: int
    base_attack_odds_change: int
    attack_odds_change_rand: int
    worse_rank_difference_attitude: int
    better_rank_difference_attitude: int
    close_borders_attitude_change: int
    lost_war_attitude_change: int
    vassal_power_modifier: int
    freedom_appreciation: int

    @classmethod
    def from_dict(cls, d: dict) -> "Leader":
        """Create a new Leader object from a dictionary."""
        d["type"] = LeaderHeadType[d["type"]]
        d["favorite_civic"] = CivicType[d["favorite_civic"]]
        d["favorite_religion"] = ReligionType[d["favorite_religion"]]
        traits = sorted([TraitType[t] for t in d["traits"]], key=lambda x: x.name)
        d["traits"] = traits[0], traits[1]
        return cls(**d)


def leader_attributes() -> List[str]:
    """Returns List of all Leader attributes."""
    return List(Leader.__dict__["__dataclass_fields__"].keys())


@lru_cache(1)
def _get_leaders_from_xml_file(xml_file: str | Path) -> Dict[str, Leader]:
    leaders = {}

    text_map_json = (
        importlib_resources.files("civ4save.contrib.data")
        .joinpath("text_map.json")
        .read_text()
    )
    text_map = json.loads(text_map_json)

    with open(xml_file, "r") as f:
        data = xmltodict.parse(f.read())

    leader_infos = data["Civ4LeaderHeadInfos"]["LeaderHeadInfos"]["LeaderHeadInfo"]
    for leader_info in leader_infos:
        type = LeaderHeadType[leader_info["Type"]]
        if type == LeaderHeadType.LEADER_BARBARIAN:
            continue
        description = text_map.get(
            leader_info["Description"], leader_info["Description"]
        )

        try:
            favorite_civic = CivicType[leader_info["FavoriteCivic"]]
        except KeyError:
            favorite_civic = CivicType.NO_CIVIC
        try:
            favorite_religion = ReligionType[leader_info["FavoriteReligion"]]
        except KeyError:
            favorite_religion = ReligionType.NO_RELIGION

        trait_types = sorted(
            [TraitType[t["TraitType"]] for t in leader_info["Traits"]["Trait"]],
            key=lambda x: x.name,
        )
        traits = trait_types[0], trait_types[1]
        try:
            flavors = {
                f["FlavorType"]: int(f["iFlavor"])
                for f in leader_info["Flavors"]["Flavor"]
            }
        except TypeError:
            flavors = {}

        try:
            u_mods = leader_info["UnitAIWeightModifiers"]["UnitAIWeightModifier"]
            u_type = UnitAIType[u_mods["UnitAIType"]]
            u_weight = int(u_mods["iWeightModifier"])
            unit_ai_weight_modifiers = {u_type.name: u_weight}
        except TypeError:
            unit_ai_weight_modifiers = {}

        try:
            i_mods = leader_info["ImprovementWeightModifiers"][
                "ImprovementWeightModifier"
            ]
            i_type = ImprovementType[i_mods["ImprovementType"]]
            i_weight = int(i_mods["iWeightModifier"])
            improvement_weight_modifiers = {i_type.name: i_weight}
        except TypeError:
            improvement_weight_modifiers = {}

        wonder_construct_rand = int(leader_info["iWonderConstructRand"])
        base_attitude = int(leader_info["iBaseAttitude"])
        base_peace_weight = int(leader_info["iBasePeaceWeight"])
        peace_weight_rand = int(leader_info["iPeaceWeightRand"])
        warmonger_respect = int(leader_info["iWarmongerRespect"])
        espionage_weight = int(leader_info["iEspionageWeight"])
        refuse_to_talk_war_threshold = int(leader_info["iRefuseToTalkWarThreshold"])
        no_tech_trade_threshold = int(leader_info["iNoTechTradeThreshold"])
        tech_trade_known_percent = int(leader_info["iTechTradeKnownPercent"])
        max_gold_trade_percent = int(leader_info["iMaxGoldTradePercent"])
        max_gold_trade_per_turn_percent = int(
            leader_info["iMaxGoldPerTurnTradePercent"]
        )
        max_war_rand = int(leader_info["iMaxWarRand"])
        max_war_nearby_power_ratio = int(leader_info["iMaxWarNearbyPowerRatio"])
        max_war_distant_power_ratio = int(leader_info["iMaxWarDistantPowerRatio"])
        max_war_min_adjacent_land_percent = int(
            leader_info["iMaxWarMinAdjacentLandPercent"]
        )
        limited_war_rand = int(leader_info["iLimitedWarRand"])
        limited_war_power_ratio = int(leader_info["iLimitedWarPowerRatio"])
        dogpile_war_rand = int(leader_info["iDogpileWarRand"])
        make_peace_rand = int(leader_info["iMakePeaceRand"])
        declare_war_trade_rand = int(leader_info["iDeclareWarTradeRand"])
        demand_rebuked_sneak_prob = int(leader_info["iDemandRebukedSneakProb"])
        demand_rebuked_war_prob = int(leader_info["iDemandRebukedWarProb"])
        raze_city_prob = int(leader_info["iRazeCityProb"])
        build_unit_prob = int(leader_info["iBuildUnitProb"])
        base_attack_odds_change = int(leader_info["iBaseAttackOddsChange"])
        attack_odds_change_rand = int(leader_info["iAttackOddsChangeRand"])
        worse_rank_difference_attitude = int(
            leader_info["iWorseRankDifferenceAttitudeChange"]
        )
        better_rank_difference_attitude = int(
            leader_info["iBetterRankDifferenceAttitudeChange"]
        )
        close_borders_attitude_change = int(leader_info["iCloseBordersAttitudeChange"])
        lost_war_attitude_change = int(leader_info["iLostWarAttitudeChange"])
        vassal_power_modifier = int(leader_info["iVassalPowerModifier"])
        freedom_appreciation = int(leader_info["iFreedomAppreciation"])

        leader = Leader(
            type=type,
            description=description,
            favorite_civic=favorite_civic,
            favorite_religion=favorite_religion,
            traits=traits,
            flavors=flavors,
            unit_ai_weight_modifiers=unit_ai_weight_modifiers,
            improvement_weight_modifiers=improvement_weight_modifiers,
            wonder_construct_rand=wonder_construct_rand,
            base_attitude=base_attitude,
            base_peace_weight=base_peace_weight,
            peace_weight_rand=peace_weight_rand,
            warmonger_respect=warmonger_respect,
            espionage_weight=espionage_weight,
            refuse_to_talk_war_threshold=refuse_to_talk_war_threshold,
            no_tech_trade_threshold=no_tech_trade_threshold,
            tech_trade_known_percent=tech_trade_known_percent,
            max_gold_trade_percent=max_gold_trade_percent,
            max_gold_trade_per_turn_percent=max_gold_trade_per_turn_percent,
            max_war_rand=max_war_rand,
            max_war_nearby_power_ratio=max_war_nearby_power_ratio,
            max_war_distant_power_ratio=max_war_distant_power_ratio,
            max_war_min_adjacent_land_percent=max_war_min_adjacent_land_percent,
            limited_war_rand=limited_war_rand,
            limited_war_power_ratio=limited_war_power_ratio,
            dogpile_war_rand=dogpile_war_rand,
            make_peace_rand=make_peace_rand,
            declare_war_trade_rand=declare_war_trade_rand,
            demand_rebuked_sneak_prob=demand_rebuked_sneak_prob,
            demand_rebuked_war_prob=demand_rebuked_war_prob,
            raze_city_prob=raze_city_prob,
            build_unit_prob=build_unit_prob,
            base_attack_odds_change=base_attack_odds_change,
            attack_odds_change_rand=attack_odds_change_rand,
            worse_rank_difference_attitude=worse_rank_difference_attitude,
            better_rank_difference_attitude=better_rank_difference_attitude,
            close_borders_attitude_change=close_borders_attitude_change,
            lost_war_attitude_change=lost_war_attitude_change,
            vassal_power_modifier=vassal_power_modifier,
            freedom_appreciation=freedom_appreciation,
        )
        leaders[description] = leader
    return leaders


@lru_cache(1)
def _get_leader_map():
    """Load from json file in contrib/data."""
    leaders_json = (
        importlib_resources.files("civ4save.contrib.data")
        .joinpath("leaders.json")
        .read_text()
    )
    leaders = json.loads(leaders_json)
    return leaders


def rank_leaders(attribute: str, reverse: bool = False) -> List[Tuple]:
    """Returns a sorted List of Tuples (description, attribute) by attribute value.

    Accepts attributes in CamelCase or snake_case.
    """
    leader_map = _get_leader_map()
    leaders = [Leader.from_dict(leader_map[ld]) for ld in leader_map]

    # convert Camel -> snake
    attribute = "".join(
        ["_" + c.lower() if c.isupper() else c for c in attribute]
    ).lstrip("_")

    ranked: List[Leader] = []
    if attribute == "traits":
        ranked = sorted(leaders, key=lambda x: x.traits[0].name)

    while not ranked:
        try:
            ranked = sorted(
                leaders, key=lambda x: getattr(x, attribute), reverse=reverse
            )
        except TypeError:
            try:
                # Enum
                ranked = sorted(
                    leaders, key=lambda x: getattr(x, attribute).name, reverse=reverse
                )
            except AttributeError:
                # Dict
                ranked = sorted(leaders, key=lambda x: x.description)

    return [(ld.description, getattr(ld, attribute)) for ld in ranked]


def get_leader(name: str) -> Optional[Leader]:
    """Returns Leader where name == Leader.description or None."""
    leaders = _get_leader_map()
    l_dict = leaders.get(name, None)
    if l_dict:
        return Leader.from_dict(l_dict)
    return None


def get_leaders() -> Dict[str, Leader]:
    """Returns a dict of leader descriptions mapped to the Leader object."""
    leaders = _get_leader_map()
    return {ld: Leader.from_dict(leaders[ld]) for ld in leaders}


if __name__ == "__main__":
    # from ..utils import CustomJsonEncoder
    # LEADER_INFOS = Path("xml") / "CIV4LeaderHeadInfos.xml"
    # leaders = _get_leaders_from_xml_file(LEADER_INFOS)
    # print(json.dumps(leaders, cls=CustomJsonEncoder, indent=4))
    print(leader_attributes())



"""Public API for civ4save.objects."""
from .context import Context  # noqa: F401
from .game_state import GameState  # noqa: F401
from .player import Player, get_players  # noqa: F401
from .plot import Plot  # noqa: F401
from .settings import Settings  # noqa: F401
